"""
OAuth2 provider customized `django-oauth2-provider` forms.

"""

from django.contrib.auth import authenticate
from django.contrib.auth.models import User

import provider.oauth2.forms
import provider.constants
from provider.forms import OAuthValidationError
from provider.oauth2.forms import ScopeChoiceField
from provider.oauth2.models import Client

from .constants import SCOPE_NAMES


# The following forms override the scope field to use the SCOPE_NAMES
# defined for this provider. Otherwise it will use the default values from
# the django-oauth2-provider package.

# pylint: disable=missing-docstring,no-member

class AuthorizationRequestForm(provider.oauth2.forms.AuthorizationRequestForm):
    def __init__(self, *args, **kwargs):
        super(AuthorizationRequestForm, self).__init__(*args, **kwargs)
        self.fields['scope'] = ScopeChoiceField(choices=SCOPE_NAMES, required=False)


class AuthorizationForm(provider.oauth2.forms.AuthorizationForm):
    def __init__(self, *args, **kwargs):
        super(AuthorizationForm, self).__init__(*args, **kwargs)
        self.fields['scope'] = ScopeChoiceField(choices=SCOPE_NAMES, required=False)


class RefreshTokenGrantForm(provider.oauth2.forms.RefreshTokenGrantForm):
    def __init__(self, *args, **kwargs):
        super(RefreshTokenGrantForm, self).__init__(*args, **kwargs)
        self.fields['scope'] = ScopeChoiceField(choices=SCOPE_NAMES, required=False)


class AuthorizationCodeGrantForm(provider.oauth2.forms.AuthorizationCodeGrantForm):
    def __init__(self, *args, **kwargs):
        super(AuthorizationCodeGrantForm, self).__init__(*args, **kwargs)
        self.fields['scope'] = ScopeChoiceField(choices=SCOPE_NAMES, required=False)


# pylint: enable=missing-docstring,no-member

# The forms in this module are required to use email as a secondary
# identifier when authenticating via OAuth2, since the specification
# only uses the `username` parameter.
#
# TODO: An alternative, simpler, approach is to write a backend that
# like `django.contrib.auth.backends.ModelBackend` and add it to
# `AUTHENTICATION_BACKENDS` in the Django settings.


class PasswordGrantForm(provider.oauth2.forms.PasswordGrantForm):
    """
    Forms that validates the user email to be used as secondary user
    identifier during authentication.

    """

    def clean(self):
        data = self.cleaned_data  # pylint: disable=no-member
        username = data.get('username')
        password = data.get('password')

        user = authenticate(username=username, password=password)

        # If the username was not found try the user using username as
        # the email address. It is valid because the edx-platform has
        # a unique constraint placed on the email field.
        if user is None:
            try:
                user_obj = User.objects.get(email=username)
                user = authenticate(username=user_obj.username, password=password)
            except User.DoesNotExist:
                user = None

        if (
            user is None
            # TODO This is a temporary workaround while the is_active field on the
            # user is coupled with whether or not the user has verified ownership
            # of their claimed email address.  Once is_active is decoupled from
            # verified_email, we can uncomment the following line.
            # or not user.is_active
        ):
            raise OAuthValidationError({'error': 'invalid_grant'})

        data['user'] = user
        return data


class PublicPasswordGrantForm(PasswordGrantForm,
                              provider.oauth2.forms.PublicPasswordGrantForm):
    """
    Form wrapper to ensure the the customized PasswordGrantForm is used
    during client authentication.

    """
    def clean(self):
        data = super(PublicPasswordGrantForm, self).clean()

        try:
            client = Client.objects.get(client_id=data.get('client_id'))
        except Client.DoesNotExist:
            raise OAuthValidationError({'error': 'invalid_client'})

        if client.client_type != provider.constants.PUBLIC:
            raise OAuthValidationError({'error': 'invalid_client'})

        data['client'] = client
        return data



"""
OAuth2 provider urls
"""

from django.conf.urls import patterns, url
from django.contrib.auth.decorators import login_required
from django.views.decorators.csrf import csrf_exempt
from provider.oauth2.views import AccessTokenDetailView

from .views import Authorize, Redirect, Capture, AccessTokenView, UserInfoView


urlpatterns = patterns(
    '',
    url('^authorize/?$', login_required(Capture.as_view()), name='capture'),
    url('^authorize/confirm/?$', login_required(Authorize.as_view()), name='authorize'),
    url('^redirect/?$', login_required(Redirect.as_view()), name='redirect'),
    url('^access_token/?$', csrf_exempt(AccessTokenView.as_view()), name='access_token'),
    url('^access_token/(?P<token>[\w]+)/$', csrf_exempt(AccessTokenDetailView.as_view()), name='access_token_detail'),
    url('^user_info/?$', csrf_exempt(UserInfoView.as_view()), name='user_info'),
)



import json
from optparse import make_option

from django.core.exceptions import ValidationError
from django.core.management.base import BaseCommand, CommandError
from django.core.validators import URLValidator
from provider.constants import CONFIDENTIAL, PUBLIC
from provider.oauth2.models import Client

from ...models import TrustedClient

try:
    from django.contrib.auth import get_user_model
except ImportError:  # Django <1.5
    from django.contrib.auth.models import User
else:
    User = get_user_model()

ARG_STRING = '<url> <redirect_uri> <client_type: "confidential" | "public">'


class Command(BaseCommand):
    help = 'Create a new OAuth2 Client. Outputs a serialized representation of the newly-created Client.'
    args = ARG_STRING
    fields = None

    option_list = BaseCommand.option_list + (
        make_option(
            '-u',
            '--username',
            action='store',
            type='string',
            dest='username',
            help="Username of a user to associate with the Client."
        ),
        make_option(
            '-n',
            '--client_name',
            action='store',
            type='string',
            dest='client_name',
            help="String to assign as the Client name."
        ),
        make_option(
            '-i',
            '--client_id',
            action='store',
            type='string',
            dest='client_id',
            help="String to assign as the Client ID."
        ),
        make_option(
            '-s',
            '--client_secret',
            action='store',
            type='string',
            dest='client_secret',
            help="String to assign as the Client secret. Should not be shared."
        ),
        make_option(
            '-t',
            '--trusted',
            action='store_true',
            dest='trusted',
            default=False,
            help="Designate the Client as trusted. Trusted Clients bypass the user consent "
                 "form typically displayed after validating the user's credentials."
        ),
        make_option(
            '--logout_uri',
            action='store',
            type='string',
            dest='logout_uri',
            help="Client logout URI. This value will be used for single sign out."
        ),
    )

    def handle(self, *args, **options):
        self._clean_args(args)
        self._parse_options(options)

        client_id = self.fields.get('client_id')
        trusted = self.fields.pop('trusted')

        # Check if client ID is already in use. If so, fetch existing Client and update fields.
        client_id_claimed = Client.objects.filter(client_id=client_id).exists()
        if client_id_claimed:
            client = Client.objects.get(client_id=client_id)

            for key, value in self.fields.items():
                setattr(client, key, value)

            client.save()
        else:
            client = Client.objects.create(**self.fields)

        if trusted:
            TrustedClient.objects.get_or_create(client=client)
        else:
            try:
                TrustedClient.objects.get(client=client).delete()
            except TrustedClient.DoesNotExist:
                pass

        serialized = json.dumps(client.serialize(), indent=4)
        self.stdout.write(serialized)

    def _clean_args(self, args):
        """Validate and clean the command's arguments.

        These arguments must include the Client application's URL, the Client application's
        OAuth2 callback URL, and the Client's type, indicating whether the Client application
        is capable of maintaining the confidentiality of its credentials (e.g., running on a
        secure server) or is incapable of doing so (e.g., running in a browser).

        Arguments:
            args (tuple): Arguments with which the command was called.

        Raises:
            CommandError, if the number of arguments provided is invalid, if the URLs provided
                are invalid, or if the client type provided is invalid.
        """
        if len(args) != 3:
            raise CommandError(
                "Number of arguments provided is invalid. "
                "This command requires the following arguments: {}.".format(ARG_STRING)
            )

        url, redirect_uri, client_type = args

        # Validate URLs
        for u in (url, redirect_uri):
            try:
                URLValidator()(u)
            except ValidationError:
                raise CommandError("URLs provided are invalid. Please provide valid application and redirect URLs.")

        # Validate and map client type to the appropriate django-oauth2-provider constant
        client_type = client_type.lower()
        client_type = {
            'confidential': CONFIDENTIAL,
            'public': PUBLIC
        }.get(client_type)

        if client_type is None:
            raise CommandError("Client type provided is invalid. Please use one of 'confidential' or 'public'.")

        self.fields = {
            'url': url,
            'redirect_uri': redirect_uri,
            'client_type': client_type,
        }

    def _parse_options(self, options):
        """Parse the command's options.

        Arguments:
            options (dict): Options with which the command was called.

        Raises:
            CommandError, if a user matching the provided username does not exist.
        """
        for key in ('username', 'client_name', 'client_id', 'client_secret', 'trusted', 'logout_uri'):
            value = options.get(key)
            if value is not None:
                self.fields[key] = value

        username = self.fields.pop('username', None)
        if username is not None:
            try:
                self.fields['user'] = User.objects.get(username=username)
            except User.DoesNotExist:
                raise CommandError("User matching the provided username does not exist.")

        # The keyword argument 'name' conflicts with that of `call_command()`. We instead
        # use 'client_name' up to this point, then swap it out for the expected field, 'name'.
        client_name = self.fields.pop('client_name', None)
        if client_name is not None:
            self.fields['name'] = client_name

        logout_uri = self.fields.get('logout_uri')

        if logout_uri:
            try:
                URLValidator()(logout_uri)
            except ValidationError:
                raise CommandError("The logout_uri is invalid.")



"""
OpenID Connect core related utility functions.

Defines utility functions to process the ID Token and UserInfo
endpoints according to the OpenID Connect specification.

"""

import jwt

import provider.scope

from .. import constants
from ..utils import import_string
from .collect import collect


HANDLERS = {
    'id_token': [import_string(cls) for cls in constants.ID_TOKEN_HANDLERS],
    'userinfo': [import_string(cls) for cls in constants.USERINFO_HANDLERS]
}


class IDToken(object):
    """
    Simple container for OpenID Connect related responses.

    Attributes:
        access_token (:class:`AccessToken`): Associated Access Token object.
        scopes (list): List of scope names.
        claims (dict): Dictionary of claim names and values.

    """

    def __init__(self, access_token, scopes, claims):
        self.access_token = access_token
        self.scopes = scopes
        self.claims = claims

    def encode(self, secret, algorithm='HS256'):
        """
        Encode the set of claims to the JWT (JSON Web Token) format
        according to the OpenID Connect specification:

        http://openid.net/specs/openid-connect-basic-1_0.html#IDToken

        Arguments:
            claims (dict): A dictionary with the OpenID Connect claims.
            secret (str): Secret used to encode the id_token.
            algorithm (str): Algorithm used for encoding.
                Defaults to HS256.

        Returns encoded JWT token string.

        """

        return jwt.encode(self.claims, secret, algorithm)


def id_token(access_token, nonce=None, claims_request=None):
    """
    Returns data required for an OpenID Connect ID Token according to:

    - http://openid.net/specs/openid-connect-basic-1_0.html#IDToken

    Arguments:
        access_token (:class:`AccessToken`): Associated OAuth2 access token.
        nonce (str): Optional nonce to protect against replay attacks.
        claims_request (dict): Optional dictionary with the claims request parameters.

    Information on the `claims_request` parameter specification:

    - http://openid.net/specs/openid-connect-core-1_0.html#ClaimsParameter

    Returns an :class:`IDToken` instance with the scopes from the
    access_token and the corresponding claims. Claims in the
    `claims_request` paramater id_token section will be included *in
    addition* to the ones corresponding to the scopes specified in the
    `access_token`.

    """

    handlers = HANDLERS['id_token']

    # Select only the relevant section of the claims request.
    claims_request_section = claims_request.get('id_token', {}) if claims_request else {}

    scope_request = provider.scope.to_names(access_token.scope)

    if nonce:
        claims_request_section.update({'nonce': {'value': nonce}})

    scopes, claims = collect(
        handlers,
        access_token,
        scope_request=scope_request,
        claims_request=claims_request_section,
    )

    return IDToken(access_token, scopes, claims)


def userinfo(access_token, scope_request=None, claims_request=None):
    """
    Returns data required for an OpenID Connect UserInfo response, according to:

    http://openid.net/specs/openid-connect-basic-1_0.html#UserInfoResponse

    Supports scope and claims request parameter as described in:

    - http://openid.net/specs/openid-connect-core-1_0.html#ScopeClaims
    - http://openid.net/specs/openid-connect-core-1_0.html#ClaimsParameter

    Arguments: access_token (:class:`AccessToken`): Associated access
        token.  scope_request (list): Optional list of requested
        scopes. Only scopes authorized in the `access_token` will be
            considered.  claims_request
        (dict): Optional dictionary with a claims request parameter.

    Information on the claims request parameter specification:

    - http://openid.net/specs/openid-connect-core-1_0.html#ClaimsParameter

    As a convinience, if neither `scope_request` or user_info claim is
    specified in the `claims_request`, it will return the claims for
    all the scopes in the `access_token`.

    Returns an :class:`IDToken` instance with the scopes from the
    `scope_request` and the corresponding claims. Claims in the
    `claims_request` paramater userinfo section will be included *in
    addition* to the ones corresponding to `scope_request`.

    """

    handlers = HANDLERS['userinfo']

    # Select only the relevant section of the claims request.
    claims_request_section = claims_request.get('userinfo', {}) if claims_request else {}

    # If nothing is requested, return the claims for the scopes in the access token.
    if not scope_request and not claims_request_section:
        scope_request = provider.scope.to_names(access_token.scope)
    else:
        scope_request = scope_request

    scopes, claims = collect(
        handlers,
        access_token,
        scope_request=scope_request,
        claims_request=claims_request_section,
    )

    return IDToken(access_token, scopes, claims)



""" Misc Utilities. """

from importlib import import_module
import sys

from django.utils import six


# NOTE: be careful not to create an import loop when extending this
# file, corrently `import_string` is used by `oauth2_provider.constants`.


def import_string(dotted_path):
    """
    Import a dotted module path and return the attribute/class designated by the
    last name in the path. Raise ImportError if the import failed.

    Borrowed from Django 1.7
    """
    try:
        module_path, class_name = dotted_path.rsplit('.', 1)
    except ValueError:
        msg = "%s doesn't look like a module path" % dotted_path
        six.reraise(ImportError, ImportError(msg), sys.exc_info()[2])

    module = import_module(module_path)

    try:
        return getattr(module, class_name)
    except AttributeError:
        msg = 'Module "%s" does not define a "%s" attribute/class' % (
            dotted_path, class_name)
        six.reraise(ImportError, ImportError(msg), sys.exc_info()[2])
